#include "camera.h"
#include "scene.h"
#include <qmessagebox.h>
#include "RTConfig.h"

#define GL_BGRA 0x80E1
#define GL_TEXTURE_RECTANGLE_EXT 0x84F5

Camera::Camera()
{
  setXResolution(64);
  setYResolution(64);
  setFieldOfView(M_PI / 4.0);

  //Distance between the "screen of view" and the center
  dist_ = 2;
}

void Camera::initFromDOMElement(const QDomElement& e)
{
  if (e.tagName() != "Camera")
    {
      QMessageBox::critical(NULL, "Camera init error", "Camera::initFromDOMElement, bad DOM tagName.\nExpecting 'Camera', got "+e.tagName());
      return;
    }

  RTConfig::getInstance()->setNumPrimaryRays(e.attribute("sampling", "1").toInt());

  if (!e.hasAttribute("fieldOfView"))
    QMessageBox::warning(NULL, "Camera error", "Camera has undefined fieldOfView. Using pi/4.");
  setFieldOfView(e.attribute("fieldOfView", "0.7854").toFloat());

  if (!e.hasAttribute("xResolution"))
    QMessageBox::warning(NULL, "Camera error", "Camera has undefined xResolution. Using 64.");
  setXResolution(e.attribute("xResolution", "64").toInt());

  if (!e.hasAttribute("yResolution"))
    QMessageBox::warning(NULL, "Camera error", "Camera has undefined yResolution. Using 64.");
  setYResolution(e.attribute("yResolution", "64").toInt());

  QDomNode n = e.firstChild();
  while (!n.isNull())
    {
      QDomElement e = n.toElement();
      if (!e.isNull())
	{
	  if (e.tagName() == "Frame")
	    frame_.initFromDOMElement(e);
	  else
	    QMessageBox::warning(NULL, "Camera child error", "Unsupported Camera child : "+e.tagName());
	}
      else
	QMessageBox::warning(NULL, "Camera XML error", "Error while parsing Camera XML document");
      n = n.nextSibling();
    }
}

void Camera::setXResolution(int xResolution)
{
  if (xResolution <= 0)
    {
      QMessageBox::warning(NULL, "Camera error", "Negative image xResolution - Skipping");
      return;
    }
  xResolution_ = xResolution;

  compGeometry();
}

void Camera::setYResolution(int yResolution)
{
  if (yResolution <= 0)
    {
      QMessageBox::warning(NULL, "Camera error", "Negative image yResolution - Skipping");
      return;
    }
  yResolution_ = yResolution;

  compGeometry();

}

void Camera::drawCamera(float halfHeight, float halfWidth, float dist) const {

  glPushAttrib(GL_ALL_ATTRIB_BITS);
  glDisable(GL_LIGHTING);
  glLineWidth(2.0f);
  

  const float arrowHeight    = 1.5f * halfHeight;
  const float baseHeight     = 1.2f * halfHeight;
  const float arrowHalfWidth = 0.5f * halfWidth;
  const float baseHalfWidth  = 0.3f * halfWidth;

  glColor4f(0.8,0.8,0.8,1.0);

  glEnable(GL_DEPTH_TEST);
  glDisable(GL_STENCIL_TEST);


  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  // Draw the Frustum
  //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
  glBegin(GL_LINE_STRIP);
  glVertex3f(-halfWidth, halfHeight,-dist);
  glVertex3f(-halfWidth,-halfHeight,-dist);
  glVertex3f( 0.0f, 0.0f, 0.0f);
  glVertex3f( halfWidth,-halfHeight,-dist);
  glVertex3f(-halfWidth,-halfHeight,-dist);
  glEnd();
  glBegin(GL_LINE_STRIP);
  glVertex3f( halfWidth,-halfHeight,-dist);
  glVertex3f( halfWidth, halfHeight,-dist);
  glVertex3f( 0.0f, 0.0f, 0.0f);
  glVertex3f(-halfWidth, halfHeight,-dist);
  glVertex3f( halfWidth, halfHeight,-dist);
  glEnd();

  // Draw the arrow (Up)
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
  // Base
  glBegin(GL_QUADS);
  glVertex3f(-baseHalfWidth, halfHeight,-dist);
  glVertex3f( baseHalfWidth, halfHeight,-dist);
  glVertex3f( baseHalfWidth, baseHeight,-dist);
  glVertex3f(-baseHalfWidth, baseHeight,-dist);
  glEnd();
  // Arrow
  glBegin(GL_TRIANGLES);
  glVertex3f( 0.0f,           arrowHeight,-dist);
  glVertex3f(-arrowHalfWidth, baseHeight, -dist);
  glVertex3f( arrowHalfWidth, baseHeight, -dist);
  glEnd();


  // Draw the quad with transparency (we put alpha = 0.2)

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glDisable(GL_CULL_FACE);
  glColor4f(0.15,0.94,0.65,0.2);

  //glDisable(GL_DEPTH_TEST);

  //We keep the DepthMask for a best view (so we don't see the grid of the world frame)
  // glDepthMask(false);

  //Previex texture display
  if(RTConfig::getInstance()->previewTexID && RTConfig::getInstance()->getPreviewMode()==RTConfig::RTC_PM_CAMERA ){
	glBindTexture(GL_TEXTURE_RECTANGLE_EXT, RTConfig::getInstance()->previewTexID);
	glEnable(GL_TEXTURE_RECTANGLE_EXT);

	glColor4f(1,1,1,0.7);
  }else
	  glDisable(GL_TEXTURE_RECTANGLE_EXT);


  glBegin(GL_QUADS);
	  glTexCoord2f(0,yResolution());
	  glVertex3f( -halfWidth, -halfHeight, -dist);

	  glTexCoord2f(xResolution(),yResolution());
	  glVertex3f( halfWidth, -halfHeight, -dist);

	  glTexCoord2f(xResolution(), 0);
	  glVertex3f( halfWidth, halfHeight, -dist);

	  glTexCoord2f(0, 0);
	  glVertex3f( -halfWidth, halfHeight, -dist);
  glEnd();

  glDisable(GL_BLEND);

  glEnable(GL_DEPTH_TEST);
  // Draw the lines of the grid in the quad
  glColor4f(0.7,0.7,0.7,1.0);
  glBegin(GL_LINES);

  float step=0.5f;

  // vertical lines
  for(float i=0; i<1.0; i+=step){
	glVertex3f( i*halfWidth, halfHeight, -dist);
	glVertex3f( i*halfWidth, -halfHeight, -dist);
	glVertex3f( i*-halfWidth, halfHeight, -dist);
	glVertex3f( i*-halfWidth, -halfHeight, -dist);
  }

  // horizontal lines
  for(float i=0; i<1.0; i+=step){
	glVertex3f( halfWidth, i*halfHeight, -dist);
	glVertex3f( -halfWidth, i*halfHeight, -dist);
	glVertex3f( halfWidth, i*-halfHeight, -dist);
	glVertex3f( -halfWidth, i*-halfHeight, -dist);
  }

  glEnd();


  glPopAttrib();
}


void Camera::draw(float radius) const  {
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glMultMatrixd(frame().matrix());
	glScalef(radius, radius, radius);
	drawCamera(halfHeight_, halfWidth_, dist_);
	glPopMatrix();
}

Ray Camera::getPixelRay(float i, float j) const
{
	float x = (i*2*halfWidth_)/xResolution() - halfWidth_;
	float y = -(j*2*halfHeight_)/yResolution() + halfHeight_;

	qglviewer::Vec oldDir;
	oldDir.setValue(x, y, -dist_);

	qglviewer::Vec start = this->frame().position();
	qglviewer::Vec newDir = this->frame().inverseTransformOf(oldDir.unit()).unit();

	Ray ray;
	ray.setStart(start);
	ray.setDirection(newDir);

	return ray;
}

// Draw the rays passing through each pixel of the scene
// Rk: to test, change the resolution!!!
void Camera::drawAllRays() const
{
  for (int i=0; i<xResolution(); ++i)
    for (int j=0; j<yResolution(); ++j)
		getPixelRay(i,j).draw();
}

// Draw all the rays passing through the center of each cell of the grid of the camera
void Camera::testAllRays() const
{
	for (int i=0; i<xResolution(); i+=xResolution()/4)
		for (int j=0; j<yResolution(); j+=yResolution()/4)
			getPixelRay(i+xResolution()/8,j+yResolution()/8).draw();
}
